/** 
  *  \file Apodize.hxx
  *  \brief Apodization of spectrum
  *
  *  \author M. Smuga-Otto <maciek@ssec.wisc.edu>
  *
  *  \version $Id: Apodize.hxx,v 1.1.2.4 2006/04/07 22:12:36 maciek Exp $
  *
  *  \par Copyright:
  *  \verbatim
  *
  *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004-2006
  *  Space Science and Engineering Center
  *  University of Wisconsin - Madison, USA
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  * 
  * \endverbatim
  */

#ifndef HXX_APODIZE
#define HXX_APODIZE

// C++ STL and BOOST includes
#include <algorithm>
#include <cmath>
#include <functional>
#include <boost/array.hpp>
#include <vector> 
#include <string>

// GIPS capsule includes
#include "CplxFFT3.hxx"
#include "Physics.h"
#include "FBF.h"

namespace gips {

        
template <typename InType, typename OutType>
class Apodize
{
public:
    struct Ports
    {
        const InType * obs_in;
        OutType * obs_out;
        Ports(const InType *obs_in_, OutType *obs_out_): 
             obs_in(obs_in_), obs_out(obs_out_) {}
    };
private:
    std::string APOD_FUN_LW_FILENAME;
    std::string APOD_FUN_MW_FILENAME;
    
    typedef boost::array<std::complex<double>, InType::static_size * 2> FFTBuffer; 
    typedef CplxFFT<FFTBuffer, FFTBuffer> MyFFT; 
    
    const Ports ports;
    const double normalization;
    FFTBuffer mirroredSpectrum, interferogram;
    std::vector<double> apodFun;
    

    MyFFT forwardFFT;
    MyFFT  backwardFFT; 

    template<typename FFT>
    static typename FFT::settings_t getFFTSettings(
                        typename FFT::direction_e d,
                        typename FFT::training_e t)
    {
        typename FFT::settings_t s;
        s.fftDirection = d;
        s.fftTraining = t;
        return s;
    }

public:
    Apodize(const Ports &p_): 
         APOD_FUN_LW_FILENAME("external/lw_apod_ifg.real8.2048"),
         APOD_FUN_MW_FILENAME("external/mw_apod_ifg.real8.4096"),
         ports(p_),
         normalization(static_cast<double>(ports.obs_in->size() * 2)),
         apodFun(ports.obs_in->size()*2, 0.0),
         forwardFFT(
                 getFFTSettings<MyFFT>( MyFFT::FORWARD, MyFFT::ESTIMATE),  
                 MyFFT::Ports(&mirroredSpectrum, &interferogram)),
         backwardFFT(
                 getFFTSettings<MyFFT>( MyFFT::BACKWARD, MyFFT::ESTIMATE),
                 MyFFT::Ports(&interferogram, &mirroredSpectrum))
    {
            FBF::load(apodFun, (ports.obs_in->size() == 1024) ? APOD_FUN_LW_FILENAME : APOD_FUN_MW_FILENAME);
    } 

    ~Apodize() {}
    void operator()() 
    {
        const InType &obsIn (*(ports.obs_in)); // avoid convoluted pointer deference syntax in subsequent client code.
        OutType &obsOut (*(ports.obs_out));

        std::copy(obsIn.begin(), obsIn.end(), mirroredSpectrum.begin());
        std::copy(obsIn.begin(), obsIn.end(), mirroredSpectrum.rbegin()); 

        forwardFFT(); // reap benefits of CplxFFT capsule

        for(unsigned int i=0; i< interferogram.size(); ++i)
                interferogram[i] *= apodFun[i];

        backwardFFT();

        for (unsigned int i=0; i< obsOut.size(); ++i)
                obsOut[i] = std::real(mirroredSpectrum[i]) / normalization;

    } // end operator()
}; // end class Apodize
} // end namespace gips
#endif

