/**
 * \file Planck.hxxh
 * Planck curve utility library
 *
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 * \version $Id: Planck.hxx,v 1.4.4.3 2006/03/30 20:22:47 graemem Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2005
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 */

#ifndef H_PLANCK
#define H_PLANCK

#include "Physics.h"

namespace gips
{

/*! Physical constants used in the calculation of blackbody radiation curves.
  *
  *  \par FUTUREs: 
  *  - Theory reference missing
  *  - How to load these constants from some external reference struct?
  */
class PlanckConstants
{
protected:
    const double h,c,k;
    const double c1, c2;
    
public:
    PlanckConstants():
            h(6.62606876E-34),
            c(2.99792458E8),
            k(1.3806503E-23),
            c1(2.0*h*c*c*1e8),
            c2(h*c/k*1e2)
    { }
    
    PlanckConstants& operator=( const PlanckConstants &x ) { return *this; }
};


/*! Compute the radiance curve for a blackbody at a given temperature.
  * This class offers both index-based access and iterator-like access.
  *
  *  \par Requirements:
  *  Class needs to be initialized with a temperature and a pointer to a wavenumber scale.
  *
  *  \par Restrictions:
  *  The wavenumber scale type must support array-like access semantics.
  */
template< typename W >
class PlanckRadiances: private PlanckConstants
{
    double T; // Kelvin
    const W *w;

public:
    PlanckRadiances( double T_ = 0.0, const W *w_ = 0 ): 
            T(T_), 
            w(w_)
    {
    }
        
    static const std::size_t static_size = W::static_size;    
    typedef double value_type;
    
    inline unsigned size() const { return w==NULL? 0 : W::static_size; }
    
    inline bool operator==( const PlanckRadiances<W> &it ) const 
    { return it.T==T && it.w==w; }
    
    inline bool operator!=( const PlanckRadiances<W> &it ) const 
    { return ! (it==*this); }

    inline double operator[]( unsigned offset ) const
    {
        if (!w) return NAN;
        const W &wavenumbers(*w);
        double wn = wavenumbers[offset];
        return 1000.0 * c1 * (wn * wn * wn) / (exp((c2 * wn) / T) - 1.0);
    }

    typedef indexed_array_const_fwd_bwd_iterator< PlanckRadiances<W> > 
        iterator, const_iterator, reverse_iterator, const_reverse_iterator;
    
    inline iterator begin() const { return iterator(*this); }
    inline iterator end() const { return iterator(*this,size()); }
    
    inline iterator rbegin() const { return iterator(*this,0,-1); }
    inline iterator rend() const { return iterator(*this,size(),-1); }

};

/*! Utility class for calculating blackbody radiance curves adjusted for known emissivity profiles.
  *
  *  \par Theory: 
  *  for any wn, the adjusted radiance for temperature t, at wavenumber wn is given by
  *           b_t(wn) * e(wn) + b_rt(wn) * (1 - e(wn))
  *  where b_t is the pure planck curve at temperature t, rt is the reflected temperature of the 
  *  surrounding optical system, and e is the emissivity.
  *
  *  \par Requirements:
  *  Besides PlanckRadiances requirements, also needs a pointer to a vector describing the
  *  emissivity of the observed blackbody cavity, 
  *  and a reflected temperature of the surrounding optical system
  *
  *  \par Restrictions:
  *  The emissivity must be given at a matching wavenumber scale as the underlying PlanckRadiance.
  *
  */
template< typename W, typename EmissVector >
class EffectivePlanckRadiances
{
    PlanckRadiances<W> bb, refl;
    const EmissVector &emissivity; // 0.0 - 1.0 emissivity fraction matched to wavenumber scale
public:
    
    // return effective planck radiance for a blackbody of a given temperature Tbb (Kelvin),
    // with a reflected temperature of the optical system observing it being Trefl,
    // and an emissivity of the cavity of emiss. 
    EffectivePlanckRadiances( double Tbb_, double Trefl_, const EmissVector *emiss_, const W *w_ ): 
        bb(Tbb_,w_), refl(Trefl_,w_), emissivity(*emiss_) 
    {
        assert( w_->size() == emiss_->size() ); 
    }
        
    static const std::size_t static_size = W::static_size;    
    typedef double value_type;
    
    inline unsigned size() const { return W::static_size; }
    
    inline bool operator==( const EffectivePlanckRadiances<W,EmissVector> &it ) const 
    { return it.bb==bb && it.refl==refl && it.emissivity==emissivity; }
    
    inline bool operator!=( const EffectivePlanckRadiances<W,EmissVector> &it ) const 
    { return ! (it==*this); }

    inline double operator[]( unsigned offset ) const
    {
        return bb[offset] * emissivity[offset] + refl[offset] * (1.0 - emissivity[offset]);
    }

    typedef indexed_array_const_fwd_bwd_iterator< EffectivePlanckRadiances<W,EmissVector> > 
        iterator, const_iterator, reverse_iterator, const_reverse_iterator;
    
    inline iterator begin() const { return iterator(*this); }
    inline iterator end() const { return iterator(*this,W::static_size); }
    
    inline iterator rbegin() const { return iterator(*this,0,-1); }
    inline iterator rend() const { return iterator(*this,W::static_size,-1); }


};

} // namespace

#endif
