/**
 * Basic C++ API for the Flat Binary Format.
 * Contains read,write and sizing utilites sufficient for GIPS purposes.
 *
 * author: Maciek Smuga-Otto
 * adapted from the C FBF library by Ray Garcia
 * Copyright 2006 UW-Madison
 * 
 * CVS version: 
 * $Id: FBF.h,v 1.4.2.1 2006/02/27 21:20:33 graemem Exp $ 
**/

#ifndef FBFH
#define FBFH

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

/* STL tools */
#include <string>

namespace FBF { 

// load a flat binary file in its entirety into a resizable STL container 
template < typename V > 
int load( V &out, FILE *fp, bool autoclose = false, size_t group_size = 16384 ) 
{ 
    typedef typename V::value_type M;
    if (!fp) return -1;
    size_t nread = 0, result = group_size;
    while( !feof(fp) && (result==group_size) ) 
    { // eat the file in chunks 
        out.resize( nread + group_size );
        result = fread( &out[nread], sizeof(M), group_size, fp );
        nread += result;
    }
    out.resize( nread );
    if (autoclose) 
        fclose(fp);
    return (nread==0)? 1 : 0;
}

// as above, but accept filename instead of file pointer 
template <typename V>
int load(V &out, const std::string &filename)
{
    int rc = load( out, fopen( filename.c_str(), "rb" ), true );
    return rc;
}

// load a (small) flat binary file in its entirety into a statically sized STL container 
// intended for debugging uses, such as reading a single spectrum into a boost::array  
// FIXME - this function seems to cut out the last record! 
template < typename V > 
int load_static( V &out, FILE *fp, bool autoclose = false) 
{ 
    typedef typename V::value_type M;
    size_t group_size = V::static_size;
    if (!fp) return -1;
    size_t nread = 0, result = group_size;
    while( !feof(fp) && (result==group_size) ) 
    { /* eat the file in chunks */
        result = fread( &out[nread], sizeof(M), group_size - 1, fp );
        //result = fread( &out[nread], sizeof(M), group_size - 1, fp );
        nread += result;
    }
    if (autoclose) 
        fclose(fp);
    return (nread==0)? 1 : 0;
}

// as above, but accepts a C++ string filename instead of file pointer 
template <typename V>
int load_static(V &out, const std::string &filename)
{
    int rc = load_static( out, fopen( filename.c_str(), "rb" ), true );
    return rc;
}

// save data from any contiguous storage container (array, boost::array, vector)
// to a named FBF file
template< typename V >
int save( const std::string& filename, const V &data, int flags=0, int mode=0664 )
{ 
    int fd = 0;
    fd = open( filename.c_str(), O_WRONLY | O_CREAT | flags, mode );
    //WHENTESTING( printf( "creating %s..\n", filename.c_str() ) );
    if (fd<0) return fd;
    ssize_t block = (ssize_t)sizeof(typename V::value_type) * data.size();
    //WHENTESTING( printf( "fd: %d..\n", fd ) );
    ssize_t writ = write( fd, &data[0], block );
    close(fd);
    //WHENTESTING( printf( "wrote %ld of %ld bytes..\n", (long)writ, (long)block ) );
    return writ==block? 0 : -1;
}

// fill a passed-in resizable container with dimensions of a named FBF file
// e.g. foo.real4.45.23 pushes back 23, then 45 onto the output container.
template< typename V >
size_t dimensions( V &out, const std::string& file )
{ /* return the dimensions on the filename in major-first order */
  /* e.g. foo.real4.45.23 -> [ 23, 45 ] */
    if (file.length() == 0)
        return 1;
    std::string::size_type idx = file.find('.');
    std::string remaining = (idx == std::string::npos)
            ? ""
            : file.substr(idx + 1);
        
    size_t nelems = dimensions<V>( out, remaining );
    if ( isdigit(remaining[0]) )
    {
        int dim = atoi(remaining.c_str());
        out.push_back( dim );
        nelems *= (size_t)dim;
    }
    return nelems;
}

} /* namespace */

#endif
