/** 
*  \file Physics.h
*  \brief contains definitions in the Physics namespace
*  \author R.K.Garcia <rayg@ssec.wisc.edu>
*
*  \version $Id: Physics.h,v 1.16.2.6 2006/03/30 20:31:00 graemem Exp $
*
*  \par Copyright:
*  \verbatim
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*  \endverbatim
*/


#ifndef H_PHYSICS
#define H_PHYSICS

#include <cmath>
#include <iterator>

// DEBUG *********
#include <iostream>
// ***************

//#define DEBUG_MODE

/**
*  \brief Physics namespace, including standard types and formulae for 
*  radiances, wavenumbers, temperatures, times, voltages.
*/
namespace gips 
{

// FUTURE: move this to a better place
    template< typename IA >
    class indexed_array_const_fwd_bwd_iterator
    {
      const IA &source;
      unsigned i;
      const int incr;
      
      public:
      indexed_array_const_fwd_bwd_iterator( const IA &source_, unsigned i_=0, int incr_=1 ): 
              source(source_), i(i_), incr(incr_)
      { }
      
      typename IA::value_type operator*() const { return source[i]; }
      indexed_array_const_fwd_bwd_iterator& operator++() { i+=incr; return *this; }
      indexed_array_const_fwd_bwd_iterator& operator++(int) { i+=incr; return *this; }
      indexed_array_const_fwd_bwd_iterator& operator--() { i-=incr; return *this; }
      indexed_array_const_fwd_bwd_iterator& operator--(int) { i-=incr; return *this; }
      
      inline bool operator==( const indexed_array_const_fwd_bwd_iterator &it ) const
      {
          return it.source==source && i==it.i;
      }
      
      inline bool operator!=( const indexed_array_const_fwd_bwd_iterator &it ) const
      {
          return it.source!=source || i!=it.i;
      }
      
      inline int operator-( const indexed_array_const_fwd_bwd_iterator &it ) const
      {
          return i - it.i;
      }
      
      typedef std::bidirectional_iterator_tag iterator_category;
      typedef typename IA::value_type value_type;
      typedef typename IA::value_type *pointer;
      typedef typename IA::value_type &reference;
      typedef int difference_type;
    };
	
    namespace physics 
    { 
	
    /**
     * square of the half-angle subtended by a detector pixel.
     */
    const double BB(0.15243531037903); 
                
	/**
	 * floating point time representation
	 * Units: seconds
	 */
	typedef double seconds_t;
	
	/**
	 * floating point temperature representation
	 * Units: Kelvin
	 */
	typedef double kelvin_t;
	
	/**
	 * floating point wavenumber representation
	 * Units: wavenumbers
	 */
	typedef double wnum_t;
	
	/** 
	 *inverse of wavenumbers - centimeters.
	 */
	typedef double cm_t;
	
	/** 
	 * Radiance_t value
	 * Units: 
	 */
	typedef double radiance_t;
	
    enum observation_type_e
    {
        OBS_INVALID=0,   /* unknown, invalid or otherwise corrupt */  
        OBS_REF_UPPER=1, /* HBB, upper reference blackbody */
        OBS_REF_LOWER=2, /* ABB/WBB, lower reference blackbody */
        OBS_REF_SPACE=3, /* SBB/ZBB, space reference */
        OBS_EARTH=4,     /* downlooking earth radiance: GIFTS, S-HIS, NAST_I, AERI */
        OBS_SKY=5,       /* uplooking aircraft instrument sky view: AERI, S-HIS */ 
        OBS_LIMB=6       /* limb sounding: TES, AERI? */
    };
    
    enum scan_direction_e
    {
        SCAN_FWD=1,     /* interferometer forward scan */
        SCAN_BWD=2      /* interferometer backward scan */
    };
    	
    	
	/** \fun planckRadiance
	 * \param T temperature in Kelvin
	 * \param wnum wavenumber
	 * \todo Implement me!
	 */
//	inline radiance_t planckRadiance( kelvin_t T, wnum_t wnum )
//	{
//	    return radiance_t( 0. ); 
//	};

    // regularly spaced wavenumbers POD for interferometer instruments
	struct wnums_t
	{
	    double baseWnum;
	    double deltaWnum;
	    unsigned wnumCount;
            
	    /** access as if it were an array */
	    inline double operator[]( unsigned offset ) const
		{ return deltaWnum*double(offset)+baseWnum ; }
				
		inline bool operator==( const wnums_t &it ) const 
		{ return it.baseWnum==baseWnum && it.deltaWnum==deltaWnum && it.wnumCount==wnumCount; }
		
		inline bool operator!=( const wnums_t &it ) const 
		{ return ! (it==*this); }		

        static wnums_t fromLaserWnum(unsigned spcSize, double vlaser, unsigned reductionFactor) 
            // from ROK, wavenumbers for spectral form; move to CDF reference file
            // call with LW_IFG_SIZE or MW_IFG_SIZE for normal use
            // nominal vlaser (effective laser wavenumber) for LW = 9398.0 per ROK code
        {        
            double xmax = double(spcSize) * double(reductionFactor)/vlaser;
            double dv = 1.0 / (2.0 * xmax);
      
            double v0 = spcSize * dv; // correct for nominal GIFTS wavenumber scales
            wnums_t ret;
            ret.baseWnum = v0;
            ret.deltaWnum = dv;
            ret.wnumCount = spcSize;
            return ret;
        }
	};
	    

    /** standardized view of regularly-spaced wavenumbers */
    template <typename VTYPE, std::size_t N >
    class WnumScale
    {
    private:
        VTYPE baseWnum, deltaWnum; 
    public:
        static const std::size_t static_size = N;        
        typedef VTYPE value_type; // needed to be std::vector-like
   
        explicit WnumScale(const wnums_t &w):
            baseWnum(w.baseWnum),
            deltaWnum(w.deltaWnum)
        {
            assert(w.wnumCount==N);
        }   
        
        // FUTURE: have not found an adequate way for this class to not have a default constructor,
        // which isn't more of a problem than the default constructor itself!
        explicit WnumScale(VTYPE start_=VTYPE(nan("")), VTYPE end_=VTYPE(nan(""))):
                baseWnum(start_), 
                deltaWnum((end_ - start_) / static_cast<double>(N - 1))
        { 
        }
        
        // Calculate the wavenumber scale for a given effective or target laser wavenumber
        // and reduction (aka decimation) factor
        static WnumScale<VTYPE,N> fromLaserWnum( double vlaser, unsigned reductionFactor )
        {
            return WnumScale<VTYPE,N>( wnums_t::fromLaserWnum( N, vlaser, reductionFactor ) );
        }

        // returns a fractional index in the array of the requested wavenumber
        double indexOf(VTYPE wavenumber) const
        {
            return (wavenumber - baseWnum)/ deltaWnum;
        }
        
        VTYPE operator[](unsigned int offset) const
        {
            return deltaWnum * double(offset) + baseWnum; 
        }

        wnums_t asPOD() const
        {
            wnums_t ret;
            ret.baseWnum = baseWnum;
            ret.deltaWnum = deltaWnum;
            ret.wnumCount = N;
            return ret;
        }

        unsigned int size() const
        {
            return N;
        }
        
        bool operator==( const WnumScale &it ) const 
        { 
            return it.baseWnum == baseWnum 
                    && it.deltaWnum == deltaWnum; 
        }

        bool operator!=( const WnumScale &it ) const 
        { 
            return ! (it == *this); 
        }

        typedef indexed_array_const_fwd_bwd_iterator<WnumScale>
        iterator, const_iterator, reverse_iterator, const_reverse_iterator;

        iterator begin() const { return iterator(*this); }
        iterator end() const { return iterator(*this,N); }

        iterator rbegin() const { return iterator(*this,0,-1); }
        iterator rend() const { return iterator(*this,N,-1); }
    };

    inline double lwnRatioFromTheta(const double theta)
    {
        return 1 + (theta * theta + BB) / 2000000;
    }
    
    } // namespace physics


} // namespace gips


#endif

