/** 
  *  \file SimCal3Common.h
  *  \brief Consts and utilities for SimCal3Test and SimCal3MpiTest
  *  \verbatim
  * USAGE
  * 
  *  \endverbatim
  *
  *  \author G. Martin <graemem@ssec.wisc.edu>
  *
  *  \version $Id: SimCal3Common.h,v 1.1.2.11 2006/03/30 21:15:24 graemem Exp $
  *  \par Copyright:
  *  \verbatim
  *
  *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2006
  *  Space Science and Engineering Center
  *  University of Wisconsin - Madison, USA
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  * 
  * \endverbatim
  *
  *
  */
#ifndef H_SIMCAL3COMMON
#define H_SIMCAL3COMMON

//#include <cmath>
#include <list>
//#include <algorithm>
#include <memory>      // auto_ptr
#include <string>
#include <boost/array.hpp>
#include <boost/filesystem/path.hpp> // boost::filesystem::path
#include "netcdfcpp.h"

#include "Physics.h"   // observation type
#include "BindCdf.hxx" // provides NetCDF data record access 
#include "SimCal3.hxx" // provides the pipeline

namespace gips 
{
    const double HBB_TEMP = 300.0;  // hot blackbody temperature
    const double ABB_TEMP = 265.0;  // warm blackbody temperature ('ambient' dates back to AERI/S-HIS instruments)
    const double ZBB_TEMP = 2.76;   // space scene temperature
    const double REFL_TEMP = 200.0; // nominal temperature of the instrument optics which may radiate into the detector
    const double TELESCOPE_TAU = 1.0;
    const double CAL_MIRROR_TAU = 1.0;

//    enum band {LW_BAND, MW_BAND};
    
    // longwave consts
    const unsigned LW_IFG_SIZE=2048;
    const unsigned LW_SPC_SIZE=1024;
//const double LW_LWN = 9372.0712277413313;
    const double LW_LWN = 9398.49609375; // Approximately, 10^7 / 1064 : Nominal GIFTS laser wavenumber.
    // above value is calculated directly from WN scale presented along with 
    // simulated TOA radiances generated by the GIFTS fast model.
    const unsigned LW_DECIMATION = 8; // aka reductionFactor, ratio of raw fringe count to actual interferogram points
    const double LW_ROLLOFF_WNS [] = {628.0, 668.0, 1100.0, 1140.0};
    
    // midwave consts
    const unsigned MW_IFG_SIZE=4096;
    const unsigned MW_SPC_SIZE=2048;
//const double MW_LWN = MW_LWN; 
    const double MW_LWN = LW_LWN;  // see derivation comment at LW_LWN
    const unsigned MW_DECIMATION = 4;
    const double MW_ROLLOFF_WNS [] = {1590.0, 1630.0, 2250.0, 2290.0};

    const unsigned EMISS_SIZE = 16384;  // FUTURE: this value is currently
    // set to a generic oversample size. Will be adjusted with input  from 
    // emissivity database working group.
    
    // detector layout
    const unsigned NUM_ROWS=128, NUM_COLS=128;
    const unsigned NUM_RECS = NUM_ROWS * NUM_COLS;
  
    typedef boost::array<double, LW_IFG_SIZE> LwIfgType;
    typedef boost::array<double, MW_IFG_SIZE> MwIfgType;
    typedef boost::array<double, LW_SPC_SIZE> LwRealSpectrumType;
    typedef boost::array<double, MW_SPC_SIZE> MwRealSpectrumType;
    typedef boost::array<std::complex<double>, LW_SPC_SIZE> LwCplxSpectrumType;
    typedef boost::array<std::complex<double>, MW_SPC_SIZE> MwCplxSpectrumType;
    typedef boost::array<double, EMISS_SIZE> EmissType;

    typedef SimCal3<LwIfgType, EmissType, LwCplxSpectrumType, LwRealSpectrumType> LwSimCal3;
    typedef SimCal3<MwIfgType, EmissType, MwCplxSpectrumType, MwRealSpectrumType> MwSimCal3;
    typedef boost::shared_ptr<FileToMemBinder> FileToMemBinder_P;
    
    // construct output filename
    std::string outputFileNameFor( 
        const boost::filesystem::path &orig,   ///< original filename for source data
        const boost::filesystem::path &target_dir,  ///< target directory to place the output file in
        const char *modifier = "", ///< modifier to append to the filename stem prior to the suffix
        const char *suffix = ".nc" ); ///< suffix to append, replacing the prior suffix
        
    // calculate theta from detector id
    // FUTURE - rough calculation of thetas in place until they're read from file!
    inline double thetaFromDetectorId( unsigned j )
        {
            const static double b = 0.39042964843750;
            return std::sqrt(std::pow(b * (2 * std::abs(64.0 - int(j/128)) - 1), 2) 
                             + std::pow(b * (2 * std::abs(64.0 - j%128) - 1 ), 2));
        }

// Output file service: create GIFTS format Netcdf file, allow variables 
// to be bound to file. 
    class CdfFileOut
        {
        private:
            std::auto_ptr<NcFile> ncf;
            GiftsPixelRecordMapping pixelMapping; 
            FileToMemBinderGroup mem2disk;
            
        public:
            const std::string filename;
            
            CdfFileOut( const std::string &filename_, unsigned lwSpcSize,
                        unsigned mwSpcSize, unsigned numRows, unsigned numCols);
            
            // bind boost array to netcdf file variable
            template<typename ArrayType> 
                void bindVar(ArrayType &arr, const std::string &varname )
                {
                    long shape[] = { 1, 1, arr.size() }; 
                    mem2disk.push_back( newCdfVectorWriter( &arr, 
                                                            ncf->get_var(varname.c_str()), 
                                                            shape, shape+3, &pixelMapping ));
                }
            
            // unbind all variables
            void unbindVars()
                {
                    mem2disk.clear();
                }
            
            // iterate file bindings to next record
            void nextRec()
                {
                    mem2disk.update( +1, SEEK_CUR );
                }
            
            // seek file bindings to specified record
            void setRec(unsigned i)
                {
                    mem2disk.update( i, SEEK_SET );
                }
         };
    
// Input file service: get singleton values from input data file, allow variables 
// to be bound to file.
// FUTURE: improve error checking for file doesn't exist, variable not present
// FUTURE: get sceneID from file and compare to scenetype from user
    class CdfFileIn
        {
        private:
            std::auto_ptr<NcFile> ncf; 
            GiftsPixelRecordMapping pixelMapping;    
            GiftsThetaRecordMapping thetaMapping;
            FileToMemBinderGroup disk2mem;   
            
        public:   // expose member vars
            // used by control logic
            physics::observation_type_e scenetype; // OBS_REF_UPPER/OBS_REF_LOWER/OBS_REF_SPACE/OBS_EARTH   
            std::string filename;
            
            // values from file
            double obsTime;
            double epochTime; // must be linear/consistent/accurate measure of time
            double telescopeTemp;
            
        private:
            
            // get value from double variable varname in file; if varname does not exist, print warning to 
            // cerr and return default
            double getDouble(const std::string &varname, double defaultVal);
            
        public:
            
            // ctor
            CdfFileIn( const std::string &filename_, 
                       physics::observation_type_e scenetype_ = physics::OBS_INVALID );
            
            // bind boost array to netcdf file variable
            template <typename ArrayType>
                void bindVar(ArrayType &arr, const std::string &varname )
                {
                    long shape[] = { 1, 1, arr.size() };
                    disk2mem.push_back( newCdfVectorReader( &arr, 
                                                            ncf->get_var(varname.c_str()), 
                                                            shape, shape+3, &pixelMapping ));
                }
            
            // FUTURE: this doesn't work - not reading anything to d. Possible bindcdf problem. Fix and uncomment.
            // bind double to netcdf file variable
            /*   void bindVar(double &d, const std::string &varname )
                {
                    disk2mem.push_back( newCdfFieldReader( &d,
                                                           ncf->get_var(varname.c_str()),
                                                           &thetaMapping) );
                                                           }*/
            // unbind all variables
            void unbindVars()
                {
                    disk2mem.clear();
                }
            
            // iterate file bindings to next record
            void nextRec()
                {
                    disk2mem.update( +1, SEEK_CUR );
                }
            
            // seek file bindings to specified record
            void setRec(unsigned i)
                {
                    disk2mem.update( i, SEEK_SET );
                }
   
        };
    
    typedef std::list< boost::shared_ptr<CdfFileIn> > CdfFileInSeq; // input file sequence
    typedef std::list< boost::shared_ptr<CdfFileOut> > CdfFileOutSeq; // output file sequence
// can't use std::auto_ptr in collections, not copy constructable
       
    // return a sequence of input files based on command line args
    CdfFileInSeq parseArgs( int argc, char *argv[] );
    
    
} // namespace gips

#endif
