/** 
*  \file SimCal3Test.cc
*  \brief Driver for SimCal3 pipeline capsule
*  \verbatim
*  OVERVIEW
*
*  This file contains code for an unthreaded executable driver for the SimCal3 
*  pipeline. Handles pipeline instantiation, file I/O and pixel sequencing. 
*
*  Constants shared with SimCal3MpiTest, the parallelized driver for SimCal3, 
*  are located in the file SimCalCommon.h.
*
*  The size of the memory-resident study cache can be limited with the constant 
*  PIXEL_GROUP_SIZE. Within a pixel group, pixels are processed for all input 
*  files, then the study cache is purged. Thus a smaller value for 
*  PIXEL_GROUP_SIZE results in a smaller memory footprint, but more file 
*  binding operations.
*
*  INTERFACE 
*  
*  From the command line:
*
*    simcal3 [filetype filename] [filetype filename] 
*
*  where filetype is one of the values:
*
*       H     - hot blackbody data cube
*       W     - warm blackbody data cube
*       S     - space blackbody data cube
*       E     - Earth observation data cube
*
*  At least one Earth cube and one of each blackbody type must be specified
*
*  OUTPUT
*
*  For each input Earth file "filename.ext" , the following netcdf output files 
*  are created in the directory "Pipeline\test":
*
*     filename_unresampled.nc   <-- unresampled output
*     filename.nc               <-- final, resampled output
*
*  \endverbatim    
*  \author G. Martin <graemem@ssec.wisc.edu>
*
*  \version $Id: SimCal3Test.cc,v 1.11.2.21 2006/03/30 16:06:49 graemem Exp $
*
*  \par Copyright:
*  \verbatim
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2006
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* FUTURE: consider parallelizing file I/O. Possibly group readers and 
* writers separately within a Binder Group and specify no sharing of 
* resources among different binders to prevent race conditions. Binders
* can themselves be multi-threaded if desired.
*
*  \endverbatim
*/

#include <cassert>
#include <string>
#include <iostream> 
#include <boost/array.hpp>

#include "Physics.h"   // provides WavenumberScale<>
#include "SimCal3Common.h" // consts, typedefs and routines common between SimCal3Test and SimCal3MpiTest

using namespace std;
using namespace gips;

const unsigned PIXEL_GROUP_SIZE = 64;  // FUTURE: set this in a compile-time environment variable.

struct TestBench
{
 
    CdfFileInSeq inFiles;
        
    // wavenumber scales
    physics::WnumScale<double, LW_SPC_SIZE> lwWavenumberScale;
    physics::WnumScale<double, MW_SPC_SIZE> mwWavenumberScale;

    // main inputs/outputs
    LwIfgType   lwInBuffer;
    LwRealSpectrumType lwOutBuffer;
    MwIfgType   mwInBuffer;
    MwRealSpectrumType mwOutBuffer;
    
    // emissivity inputs
    EmissType lwEmissBuffer, mwEmissBuffer;

    // diagnostic outputs
    LwCplxSpectrumType lwObsOutBuffer;          // cplx spectrum after ifgToSpectrum
    MwCplxSpectrumType mwObsOutBuffer;          // cplx spectrum after ifgToSpectrum
    LwRealSpectrumType lwUnresampledOutBuffer;  // real spectrum after calibration  
    MwRealSpectrumType mwUnresampledOutBuffer;  // real spectrum after calibration  

    // reference structs
    LwSimCal3::ref_t       lwReference;
    MwSimCal3::ref_t       mwReference;

    // ancillary structs
    LwSimCal3::ancillary_t lwAncillary;
    MwSimCal3::ancillary_t mwAncillary;

    // control structs
    LwSimCal3::control_t   lwControl;
    MwSimCal3::control_t   mwControl;

    // ports
    LwSimCal3::Ports       lwPorts;
    MwSimCal3::Ports       mwPorts;

    // simcal3 capsules
    LwSimCal3 lwSimcal3;
    MwSimCal3 mwSimcal3;

    TestBench(const CdfFileInSeq inFiles_):
        inFiles(inFiles_),        
        lwWavenumberScale(physics::wnums_t::fromLaserWnum(LW_SPC_SIZE, LW_LWN, LW_DECIMATION )),
        mwWavenumberScale(physics::wnums_t::fromLaserWnum(MW_SPC_SIZE, MW_LWN, MW_DECIMATION )),
        lwPorts(&lwAncillary, &lwReference, &lwControl, 
                &lwInBuffer, &lwEmissBuffer,&lwObsOutBuffer, 
                &lwUnresampledOutBuffer, &lwOutBuffer),
        mwPorts(&mwAncillary, &mwReference, &mwControl, 
                &mwInBuffer, &mwEmissBuffer,&mwObsOutBuffer, 
                &mwUnresampledOutBuffer, &mwOutBuffer),
         lwSimcal3(LwSimCal3::settings_t( lwWavenumberScale, 
                                         LW_ROLLOFF_WNS[0], LW_ROLLOFF_WNS[1], 
                                         LW_ROLLOFF_WNS[2], LW_ROLLOFF_WNS[3]), 
                  lwPorts),
         mwSimcal3(MwSimCal3::settings_t( mwWavenumberScale, 
                                         MW_ROLLOFF_WNS[0], MW_ROLLOFF_WNS[1], 
                                         MW_ROLLOFF_WNS[2], MW_ROLLOFF_WNS[3]), 
                  mwPorts)
        {
            // fill buffers with NaN-bombs
            fill(lwInBuffer.begin(), lwInBuffer.end(), 
                 oobVal<LwIfgType::value_type>());
            fill(lwOutBuffer.begin(), lwOutBuffer.end(), 
                 oobVal<LwIfgType::value_type>());
            fill(lwObsOutBuffer.begin(), lwObsOutBuffer.end(), 
                 oobVal<LwCplxSpectrumType::value_type>());
            fill(lwUnresampledOutBuffer.begin(), lwUnresampledOutBuffer.end(), 
                 oobVal<LwIfgType::value_type>());
            fill(mwInBuffer.begin(), mwInBuffer.end(), 
                 oobVal<MwIfgType::value_type>());
            fill(mwOutBuffer.begin(), mwOutBuffer.end(), 
                 oobVal<MwIfgType::value_type>());
            fill(mwObsOutBuffer.begin(), mwObsOutBuffer.end(), 
                 oobVal<MwCplxSpectrumType::value_type>());
            fill(mwUnresampledOutBuffer.begin(), mwUnresampledOutBuffer.end(), 
                 oobVal<MwIfgType::value_type>());
            
            // set reference values
            lwReference.telescopeTau = TELESCOPE_TAU; 
            lwReference.calMirrorTau = CAL_MIRROR_TAU; 
            mwReference.telescopeTau = TELESCOPE_TAU; 
            mwReference.calMirrorTau = CAL_MIRROR_TAU; 
            
            // set default ancillary values - the other ancillary values will
            // change in runtime
            // FUTURE - the below will in the future be sourced from the data files
            // for now, they stay constant for the simulation files provided.
            lwAncillary.abbTemp = ABB_TEMP;
            lwAncillary.hbbTemp = HBB_TEMP;
            lwAncillary.zbbTemp = ZBB_TEMP;
            lwAncillary.reflectedTemp = REFL_TEMP;
            lwAncillary.scanDir = physics::SCAN_FWD;
            mwAncillary.abbTemp = ABB_TEMP;
            mwAncillary.hbbTemp = HBB_TEMP;
            mwAncillary.zbbTemp = ZBB_TEMP;
            mwAncillary.reflectedTemp = REFL_TEMP;
            mwAncillary.scanDir = physics::SCAN_FWD;

            // set default control settings - will change in runtime
            lwControl.operation = LwSimCal3::PROCESS_DATA;
            mwControl.operation = MwSimCal3::PROCESS_DATA;
        }    

    // bind input buffer to input file; if input file is Earth file, also bind
    // output buffers to output files
    void bindBuffers(const CdfFileInSeq::const_iterator &itIn, 
                     const CdfFileOutSeq::const_iterator &itOut,
                     const CdfFileOutSeq::const_iterator &itUnresampledOut)
        {
            // bind input files to input buffers
            CdfFileIn &fileIn(**itIn);
            fileIn.bindVar(lwInBuffer, string("signalReal_lw"));
            fileIn.bindVar(mwInBuffer, string("signalReal_mw"));
            // fileIn.bindVar(lwReference.theta, string("Theta"));  // FUTURE: read from file after BindCdf bug is fixed
            // fileIn.bindVar(mwReference.theta, string("Theta"));
            
            // if earth file, bind output files to output buffers
            if (fileIn.scenetype==physics::OBS_EARTH)
            {          
                CdfFileOut &fileOut(**itOut);
                CdfFileOut &fileUnresampledOut(**itUnresampledOut);
                fileOut.bindVar(lwOutBuffer, string("signalReal_lw"));
                fileOut.bindVar(mwOutBuffer, string("signalReal_mw"));
                fileUnresampledOut.bindVar(lwUnresampledOutBuffer, string("signalReal_lw"));
                fileUnresampledOut.bindVar(mwUnresampledOutBuffer, string("signalReal_mw"));
                // FUTURE: output baseWnum, deltaWnum
            }              
        }

    // unbind input buffers; if Earth file, unbind output buffers
    void unbindBuffers(const CdfFileInSeq::const_iterator &itIn, 
                       const CdfFileOutSeq::const_iterator &itOut,
                       const CdfFileOutSeq::const_iterator &itUnresampledOut)
        {
            // unbind input buffers
            CdfFileIn &fileIn(**itIn);
            fileIn.unbindVars();
            
            // if earth file, unbind output buffers
            if (fileIn.scenetype==physics::OBS_EARTH)
            {          
                CdfFileOut &fileOut(**itOut);
                CdfFileOut &fileUnresampledOut(**itUnresampledOut);
                fileOut.unbindVars();
                fileUnresampledOut.unbindVars();
            }                 
        }

    void setRec(const CdfFileInSeq::const_iterator &itIn, 
                const CdfFileOutSeq::const_iterator &itOut,
                const CdfFileOutSeq::const_iterator &itUnresampledOut, 
                unsigned recNum)
        {
            (**itIn).setRec(recNum);
            if ((**itIn).scenetype==physics::OBS_EARTH)
            {          
                (**itOut).setRec(recNum); 
                (**itUnresampledOut).setRec(recNum);
            }
        }
    
    // set current record (pixel) in input file bindings; if Earth file,
    // also set current record in output file bindings
    void nextRec(const CdfFileInSeq::const_iterator &itIn, 
                 const CdfFileOutSeq::const_iterator &itOut,
                 const CdfFileOutSeq::const_iterator &itUnresampledOut)
        {
            (**itIn).nextRec();
            if ((**itIn).scenetype==physics::OBS_EARTH)
            {          
                (**itOut).nextRec(); 
                (**itUnresampledOut).nextRec();
            }
        }
    
    // iterate to next file in input sequence; if earth file, also iterate
    // to next file in output sequences
    void nextFile(CdfFileInSeq::const_iterator &itIn, 
                  CdfFileOutSeq::const_iterator &itOut,
                  CdfFileOutSeq::const_iterator &itUnresampledOut)
        {
            if ((**itIn).scenetype==physics::OBS_EARTH)
            {
                ++itOut;
                ++itUnresampledOut;
            }
            ++itIn;
        }
    // functor: process all data files, lw and mw
    void operator()()
    {
        // create output files for each earth file
        CdfFileOutSeq outFiles, unresampledOutFiles;
        for(CdfFileInSeq::const_iterator itIn=inFiles.begin(); itIn!=inFiles.end(); ++itIn)
        {
            CdfFileIn &fileIn(**itIn);
            if (fileIn.scenetype==physics::OBS_EARTH)
            {                  
                string outFileName = outputFileNameFor(fileIn.filename, "test");
                string unresampledFileName = outputFileNameFor(fileIn.filename, "test", 
                                                               "_unresampled");

                outFiles.push_back( CdfFileOutSeq::value_type( 
                                        new CdfFileOut(outFileName, LW_SPC_SIZE, MW_SPC_SIZE, 
                                                       NUM_ROWS, NUM_COLS)));
                unresampledOutFiles.push_back( CdfFileOutSeq::value_type( 
                                                   new CdfFileOut(unresampledFileName, LW_SPC_SIZE, 
                                                                  MW_SPC_SIZE, NUM_ROWS, NUM_COLS)));
            }
        }
        
        // for each pixel group
        for(unsigned i=0; i<NUM_RECS; i+=PIXEL_GROUP_SIZE)
        {
            
            cout<<"pixel group "<<i<<" to "<<i+PIXEL_GROUP_SIZE-1<<endl;

            // create iterators for output files
            CdfFileOutSeq::const_iterator itOut=outFiles.begin();       
            CdfFileOutSeq::const_iterator itUnresampledOut=unresampledOutFiles.begin();

            // for each input file
            for(CdfFileInSeq::const_iterator itIn=inFiles.begin(); 
                itIn!=inFiles.end(); 
                nextFile(itIn, itOut, itUnresampledOut))
            {
                // bind input buffers, also bind output buffers if this is an earth file
                bindBuffers(itIn, itOut, itUnresampledOut);

                // FUTURE - load emissivities from ref file
                fill(lwEmissBuffer.begin(), lwEmissBuffer.end(), 1.0);
                fill(mwEmissBuffer.begin(), mwEmissBuffer.end(), 1.0);
                
                // set ancillary data that changes per input file
                lwAncillary.obsType = (**itIn).scenetype;
                lwAncillary.time = (**itIn).epochTime;
                mwAncillary.obsType = (**itIn).scenetype;
                mwAncillary.time = (**itIn).epochTime;       
 
                // for each pixel in the pixel group 
                for(unsigned j=i; (j<i+PIXEL_GROUP_SIZE)&&(j<NUM_RECS); ++j)
                {
                    // set input record to current pixel, also set output records
                    // if this is an earth file
                    setRec(itIn, itOut, itUnresampledOut, j);

                    // FUTURE - idealized calculation of thetas in place until they're read from file
                    lwReference.theta = thetaFromDetectorId(j);
                    mwReference.theta = thetaFromDetectorId(j);
                                        
                    lwAncillary.detectorId = j;
                    mwAncillary.detectorId = j;
                    
                    lwSimcal3(); // passing in bool arg causes single writeout of FBF intermediate files
                    mwSimcal3();
                } 

                // unbind buffers
                unbindBuffers(itIn, itOut, itUnresampledOut);
            } // for each input file
            
            // pixel group has been processed: reset bb caches
            lwControl.operation = LwSimCal3::RESET_BB_CACHE;
            mwControl.operation = MwSimCal3::RESET_BB_CACHE;
            lwSimcal3();
            mwSimcal3();
            
            lwControl.operation = LwSimCal3::PROCESS_DATA;
            mwControl.operation = MwSimCal3::PROCESS_DATA;
            
        } // for each pixel group
    } // functor
}; // TestBench

// print command line usage info
void printUsageMsg(const char* argv0)
{
    std::cout 
        << "Usage: \n" 
        << "   " << argv0 << " [filetype filename] [filetype filename] ...\n" 
        << "where filetype is one of the values:\n"
        << "   H     - hot blackbody data cube\n"
        << "   W     - warm blackbody data cube\n"
        << "   S     - space blackbody data cube\n"
        << "   E     - Earth observation data cube\n"
        << "Options list must specify at least one of each blackbody and at least one observation cube.\n";
}; 

int main( int argc, char *argv[] )
{
    CdfFileInSeq inFiles; 
    inFiles = parseArgs(argc, argv);
  
    if(inFiles.size() == 0)
    {
        printUsageMsg(argv[0]);
        return 5;
    }

    static TestBench tb(inFiles);
    tb();
}

