/** 
  *  \file CdfFile.hxx
  *  Create and manipulate CDF files.
  *  Not fully tested; exception/error handling not fully implemented.
  *
  *  \author G. Martin <graemem@ssec.wisc.edu>
  *
  *  \version $Id: CdfFile.hxx,v 1.2.2.5 2006/02/27 21:20:33 graemem Exp $
  *
  *  \par Copyright:
  *  \verbatim
  *
  *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2006
  *  Space Science and Engineering Center
  *  University of Wisconsin - Madison, USA
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  * 
  * \endverbatim
  *
  * TEMPLATE INTERFACE CONTRACT:
  *
  * DoubleArrayType and ComplexArrayType must be boost-array-like types with 
  * members (as described in boost::array documentation): 
  *   begin()
  *   end()
  *   size()
  *   value_type
  * Type ComplexArrayType::value_type must have the following member functions, which
  * must return values of type DoubleArrayType::value_type (or castable):
  *   real()
  *   imag()
  * Type DoubleArray::value_type must be one of the types accepted by NetCdf's get() 
  * and put() functions (or castable). FIXME: list types
  *
  */

// FIXME - upon maturing, we should move this to Common, or IO, or something.
#ifndef HXX_CDFFILE
#define HXX_CDFFILE

#include <string>
#include <complex>
#include <functional> // required by getReal()

// NETCDF includes
#include "netcdf.hh"
#include "netcdfcpp.h"

template <typename DoubleArrayType, typename ComplexArrayType> class CdfFile
{
private:
    NcFile ncf; 
   
public:
    
    enum FileMode {
        ReadOnly = NcFile::ReadOnly,
        Write = NcFile::Write,
        Replace = NcFile::Replace,
        New = NcFile::New,
    };

    CdfFile(const char *fileName, const FileMode fm = ReadOnly) : ncf(fileName, NcFile::FileMode(fm))
        {
            assert(ncf.is_valid());
        }
 
    ~CdfFile() {}
    
    void addDim(const char *dimName, unsigned val)
        {
            NcDim *dim= ncf.add_dim(dimName, val);
            assert(dim);          
        } 
    void addDoubleVar(const char *varName, const char *dim0Name, 
                     const char *dim1Name, const char *dim2Name)
        {
            NcDim *dim0= ncf.get_dim(dim0Name);
            assert(dim0); 
            NcDim *dim1= ncf.get_dim(dim1Name);
            assert(dim1); 
            NcDim *dim2= ncf.get_dim(dim2Name);
            assert(dim2); 
            
            NcVar *var = ncf.add_var(varName,ncDouble,dim0,dim1,dim2);
            assert(var);       
        }

    void addFloatVar(const char *varName, const char *dim0Name, 
                     const char *dim1Name, const char *dim2Name)
        {
            NcDim *dim0= ncf.get_dim(dim0Name);
            assert(dim0); 
            NcDim *dim1= ncf.get_dim(dim1Name);
            assert(dim1); 
            NcDim *dim2= ncf.get_dim(dim2Name);
            assert(dim2); 
            
            NcVar *var = ncf.add_var(varName,ncFloat,dim0,dim1,dim2);
            assert(var);       
        }

    template <typename DataType>
    void writeRec(const char *varName, const DataType& data,
                  unsigned i, unsigned j) 
        {
            NcVar* var= ncf.get_var(varName);
            assert(var);
            assert(var->num_dims() == 3);
            assert(var->get_dim(0)->size() > long(i));
            assert(var->get_dim(1)->size() > long(j));
            assert(var->get_dim(2)->size() == long(data.size()));
            assert(var->set_cur(long(i), long(j), 0));
            assert(var->put(&data[0], 1, 1, long(data.size())));        
        }

    struct realExtractor: public std::unary_function<typename ComplexArrayType::value_type,
                                                     typename DoubleArrayType::value_type>
    {
         typename DoubleArrayType::value_type operator() (typename ComplexArrayType::value_type arg) const
            {
                return arg.real();
            }
    };

    struct imagExtractor: public std::unary_function<typename ComplexArrayType::value_type,
                                                     typename DoubleArrayType::value_type>
    {
        typename DoubleArrayType::value_type operator() (typename ComplexArrayType::value_type arg) const
            {
                return arg.imag();
            }
    };

    void getReal(const ComplexArrayType& data, DoubleArrayType& target) 
        { 
//            target.reserve(data.size());
//            std::transform(data.begin(), data.end(), std::back_inserter(target),             
            std::transform(data.begin(), data.end(), target.begin(), 
                           realExtractor());
        }
 
    void getImag(const ComplexArrayType& data, DoubleArrayType& target) 
        { 
            //          target.reserve(data.size());
//            std::transform(data.begin(), data.end(), std::back_inserter(target),
            std::transform(data.begin(), data.end(), target.begin(), 
                           imagExtractor());
        }
  
    void writeRec(const char *realVarName, const char *imagVarName, 
                  const ComplexArrayType& data, unsigned i, unsigned j) 
        {
            // copy real data to vector and write to file
            DoubleArrayType realData;
            getReal(data, realData);
            writeRec(realVarName, realData, i, j);

            // copy imag data to vector and copy to file
            DoubleArrayType imagData;
            getImag(data, imagData);
            writeRec(imagVarName, imagData, i, j);
        }

    template <typename DataType> 
    void readRec(const char *varName, DataType& data,
                  unsigned i, unsigned j)
        {
            NcVar* var= ncf.get_var(varName);
            assert(var);
            assert(var->num_dims() == 3);
            assert(var->get_dim(0)->size() > long(i));
            assert(var->get_dim(1)->size() > long(j));
            assert(var->get_dim(2)->size() == long(data.size()));
            assert(var->set_cur(long(i), long(j), 0));
            assert(var->get(&data[0], 1, 1, long(data.size())));        
        }
    
    void close()
        {
            ncf.close();
        }

};
#endif
