/** 
  *  \file CosineRolloffTest.cc
  *  Test CosineRolloff.hxx.
  *
  *  \author G. Martin <graemem@ssec.wisc.edu>
  *
  *  \version $Id: CosineRolloffTest.cc,v 1.2.2.8 2006/03/15 22:19:35 graemem Exp $
  *
  *  \par Copyright:
  *  \verbatim
  *
  *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2006
  *  Space Science and Engineering Center
  *  University of Wisconsin - Madison, USA
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  * 
  * \endverbatim
  *
  *
  */
//#define COSINE_ROLLOFF_STATIC_TESTING

#include <string>
#include <vector>
#include "boost/array.hpp"
#include <iostream>
#include <complex>

#include "CosineRolloff.hxx"
#include "CdfFile.hxx"

// NETCDF includes
//#include "netcdf.hh"
//#include "netcdfcpp.h"

using namespace std;
using namespace gips;

// debug level:
//   0 -- error output only (no output == success)
//   ...
//   4 -- verbose
const unsigned VERBOSITY = 0; 

const unsigned SIZE=2048;
typedef boost::array< complex< double >, SIZE > complex_vec;
typedef boost::array< double, SIZE > double_vec;
typedef boost::array< double, SIZE*2 > big_double_vec;
typedef boost::array< float, SIZE > float_vec;
typedef CdfFile< double_vec, complex_vec > CDF;

void dcout(const string& str, unsigned priority)
{
    if(priority <= VERBOSITY)
        cout<<str;
}

template <typename InType, typename OutType> struct WorkBench
{
    typedef CosineRolloff<InType, OutType> CR;
    
    typename CR::ancillary_t anc_in;
    InType obs_in;
    OutType obs_out;
    typename CR::Ports ports;
 
    CR cr;
    
    //ctor
    WorkBench():ports(&anc_in, &obs_in, &obs_out), 
                cr(ports) 
        {
            anc_in.p0 = SIZE/8-1;
            anc_in.p1 = SIZE/4-1;
            anc_in.p2 = SIZE-SIZE/4-1;
            anc_in.p3 = SIZE-SIZE/8-1;

            // no ctor for boost array
            std::fill(obs_in.begin(), obs_in.end(), typename InType::value_type(3.14));
            std::fill(obs_out.begin(), obs_out.end(), typename InType::value_type(4.14));
        }

    //ctor : pass in settings
    WorkBench(typename CR::ancillary_t& a_): anc_in(a_),
                                             ports(&anc_in, &obs_in, &obs_out), 
                                             cr(ports)
        {
            // no ctor for boost array
            std::fill(obs_in.begin(), obs_in.end(), typename InType::value_type(3.14));
            std::fill(obs_out.begin(), obs_out.end(), typename InType::value_type(4.14));
        }

    //ctor : pass in ports - buffers are external
    WorkBench(InType &obs_in_, OutType &obs_out_):
        ports(&anc_in, &obs_in, &obs_out), 
        cr(ports)
        {
            anc_in.p0 = SIZE/8-1;
            anc_in.p1 = SIZE/4-1;
            anc_in.p2 = SIZE-SIZE/4-1;
            anc_in.p3 = SIZE-SIZE/8-1;

            // no ctor for boost array
            std::fill(obs_in.begin(), obs_in.end(), typename InType::value_type(3.14));
            std::fill(obs_out.begin(), obs_out.end(), typename InType::value_type(4.14));
        }
    
};

int livelinessTest() 
{
    WorkBench<complex_vec, complex_vec> wb;
    wb.cr();
    return 0;
}

// complex data with all-zero imaginary values
int complexTest()
{
    dcout("========= complex test =========\n", 1);
    
    WorkBench<complex_vec, complex_vec> wb;
 
    // create output cdf file
    CDF outFile("complexTest.nc", CDF::Replace);
    outFile.addDim("I", 8);
    outFile.addDim("J", 8);
    outFile.addDim("wnum", SIZE);
    outFile.addDoubleVar("signalReal", "I", "J", "wnum");
    outFile.addDoubleVar("signalImag", "I", "J", "wnum");

    wb.cr();

    outFile.writeRec("signalReal", "signalImag", wb.obs_out, 0, 0);

    return 0;
}

// complex data with non-zero imaginary values
int complexTest2()
{
    dcout("========= complex test 2 =========\n", 1);
    
    WorkBench<complex_vec, complex_vec> wb;
    std::fill( wb.obs_in.begin(), wb.obs_in.end(), std::complex<double>(3.14, 11.0));

    // create output cdf file
    CDF outFile("complexTest2.nc", CDF::Replace);
    outFile.addDim("I", 8);
    outFile.addDim("J", 8);
    outFile.addDim("wnum", SIZE);
    outFile.addDoubleVar("signalReal", "I", "J", "wnum");
    outFile.addDoubleVar("signalImag", "I", "J", "wnum");

    wb.cr();

    outFile.writeRec("signalReal", "signalImag", wb.obs_out, 0, 0);

    return 0;
}

// double data
int doubleTest()
{
    dcout("========= double test =========\n", 1);
    
    WorkBench<double_vec, double_vec> wb;
 
    // create output cdf file 
    CDF outFile("doubleTest.nc", CDF::Replace);
    outFile.addDim("I", 8);
    outFile.addDim("J", 8);
    outFile.addDim("wnum", SIZE);
    outFile.addDoubleVar("signalReal", "I", "J", "wnum");

    wb.cr();

    outFile.writeRec("signalReal", wb.obs_out, 0, 0);
       
    return 0;
}

// all-zero data
int zeroTest()
{
    dcout("========= zero test =========\n", 1);

    WorkBench<double_vec, double_vec> wb;
    std::fill(wb.obs_in.begin(), wb.obs_in.end(), 0.0);
 
    // create output cdf file 
    CDF outFile("zeroTest.nc", CDF::Replace);
    outFile.addDim("I", 8);
    outFile.addDim("J", 8);
    outFile.addDim("wnum", SIZE);
    outFile.addDoubleVar("signalReal", "I", "J", "wnum");

    wb.cr();

    outFile.writeRec("signalReal", wb.obs_out, 0, 0);
       
    return 0;
}

// all-negative data
int negTest()
{
    dcout("========= neg test =========\n", 1);
    
    WorkBench<double_vec, double_vec> wb;
    std::fill(wb.obs_in.begin(), wb.obs_in.end(), -40.0);
 
    // create output cdf file 
    CDF outFile("negTest.nc", CDF::Replace);
    outFile.addDim("I", 8);
    outFile.addDim("J", 8);
    outFile.addDim("wnum", SIZE);
    outFile.addDoubleVar("signalReal", "I", "J", "wnum");

    wb.cr();

    outFile.writeRec("signalReal", wb.obs_out, 0, 0);
       
    return 0;
}

// data crosses from positive to negative
int posNegTest()
{
    dcout("========= pos/neg test =========\n", 1);

    WorkBench<double_vec, double_vec> wb;
    for(unsigned i=0; i<wb.obs_in.size(); ++i)
        wb.obs_in[i] = double(i)-1024;
 
    // create output cdf file 
    CDF outFile("posNegTest.nc", CDF::Replace);
    outFile.addDim("I", 8);
    outFile.addDim("J", 8);
    outFile.addDim("wnum", SIZE);
    outFile.addDoubleVar("signalReal", "I", "J", "wnum");

    wb.cr();

    outFile.writeRec("signalReal", wb.obs_out, 0, 0);
       
    return 0;
}

// run an entire GIFTS cube through a capsule, one pixel at a time. This is rolling
// off an interferogram, not a spectrum.
int cubeTest()
{
    dcout("========= cube test =========\n", 1);
    
    WorkBench<float_vec, float_vec> wb;

    // create input cdf file
//    CDF inFile("/home2/projects/data/gips_test/1120_1_1_ifg.cdf", CDF::ReadOnly);
    CDF inFile("../Pipeline/test/1120_hbb_1_1_ifg.cdf.cal.nc", CDF::ReadOnly);

    // create output cdf file
    CDF outFile("cubeTest.nc", CDF::Replace);
    outFile.addDim("I", 128);
    outFile.addDim("J", 128);
    outFile.addDim("wnum", SIZE);
    outFile.addFloatVar("signalReal", "I", "J", "wnum");

    for(unsigned i=0; i<128; ++i)
        for(unsigned j=0; j<128; ++j)
        {
            inFile.readRec("signalReal_lw", wb.obs_in, i, j);
            wb.cr();
            outFile.writeRec("signalReal", wb.obs_out, i, j);
        }
    
    return 0;
}

// same buffer for input and output
int oneBufferTest()
{
    dcout("========= one buffer test =========\n", 1);

    double_vec buffer;
    std::fill(buffer.begin(), buffer.end(), 3.14);
    WorkBench<double_vec, double_vec> wb(buffer, buffer);
 
    // create output cdf file 
    CDF outFile("oneBufferTest.nc", CDF::Replace);
    outFile.addDim("I", 8);
    outFile.addDim("J", 8);
    outFile.addDim("wnum", SIZE);
    outFile.addDoubleVar("signalReal", "I", "J", "wnum");

    wb.cr();

    outFile.writeRec("signalReal", wb.obs_out, 0, 0);
       
    return 0;
}

// shortest possible rolloff
int minRolloffTest()
{
    dcout("========= min rolloff test =========\n", 1);

    CosineRolloff<double_vec, double_vec>::ancillary_t anc_in;
    anc_in.p0 = 0;
    anc_in.p1 = 1;
    anc_in.p2 = SIZE - 2;
    anc_in.p3 = SIZE - 1;
    WorkBench<double_vec, double_vec> wb(anc_in);

    // create output cdf file 
    CDF outFile("minRolloffTest.nc", CDF::Replace);
    outFile.addDim("I", 8);
    outFile.addDim("J", 8);
    outFile.addDim("wnum", SIZE);
    outFile.addDoubleVar("signalReal", "I", "J", "wnum");

    wb.cr();

    outFile.writeRec("signalReal", wb.obs_out, 0, 0);
       
    return 0;
}


// complex data with all-zero imaginary values
int doubleToComplexTest()
{
    dcout("========= double to complex test =========\n", 1);
    
    WorkBench<double_vec, complex_vec> wb;
 
    // create output cdf file
    CDF outFile("doubleToComplexTest.nc", CDF::Replace);
    outFile.addDim("I", 8);
    outFile.addDim("J", 8);
    outFile.addDim("wnum", SIZE);
    outFile.addDoubleVar("signalReal", "I", "J", "wnum");
    outFile.addDoubleVar("signalImag", "I", "J", "wnum");

    wb.cr();

    outFile.writeRec("signalReal", "signalImag", wb.obs_out, 0, 0);

    return 0;
}

#ifdef COSINE_ROLLOFF_STATIC_TESTING
// fail static assert: different sizes for imput and output buffers
// just compile to test: don't need to call
int staticAssertFailTest()
{
    dcout("========= static assert fail test =========\n", 1);
    
    // should fail here
    WorkBench<double_vec, big_double_vec> wb;
/* 
    // create output cdf file 
    CDF outFile("staticAssertFailTest.nc", CDF::Replace);
    outFile.addDim("I", 8);
    outFile.addDim("J", 8);
    outFile.addDim("wnum", SIZE);
    outFile.addDoubleVar("signalReal", "I", "J", "wnum");

    wb.cr();

    outFile.writeRec("signalReal", wb.obs_out, 0, 0);
*/     
    return 0;
}
#endif

int main(int argc, char *argv[])
{
    std::vector<int (*)()> test_ptrs;
    test_ptrs.push_back(&doubleTest);          // 0
    test_ptrs.push_back(&complexTest);         // 1
    test_ptrs.push_back(&complexTest2);        // 2
    test_ptrs.push_back(&zeroTest);            // 3
    test_ptrs.push_back(&negTest);             // 4
    test_ptrs.push_back(&posNegTest);          // 5
    test_ptrs.push_back(&oneBufferTest);       // 6
    test_ptrs.push_back(&minRolloffTest);      // 7
    test_ptrs.push_back(&doubleToComplexTest); // 8
    //    test_ptrs.push_back(&cubeTest);            //
    //   test_ptrs.push_back(&staticAssertFailTest); // 
    
    if(argc == 1)
    {
        for(unsigned i = 0; i<test_ptrs.size(); ++i) 
            assert(test_ptrs[i]() == 0);
        return 0;
    }

    int arg = atoi(argv[1]);
    if(arg >= int(test_ptrs.size()))
    {
        cout<<"Error: invalid argument"<<endl;
        return 1;
    }

    assert(test_ptrs[arg]() == 0);
    return 0;

}
